-- the handler for freshly killed units
-- if we have a death animation, we need to spawn a thread to watch our position
-- and turn us into wreckage when we hit the seafloor
-- otherwise, we just use a slider

do -- mod-local block

local oldSeaUnit = SeaUnit
local oldSubUnit = SubUnit

local function CreateScaledBoom(unit, overkill, bone)
    explosion.CreateDefaultHitExplosionAtBone(
        unit,
        bone or 0,
        explosion.CreateUnitExplosionEntity(unit, overkill).Spec.BoundingXZRadius
    )
end

local function CreateWreckageProp( self, overkillRatio )
    local bp = self:GetBlueprint()
    local wreck = bp.Wreckage.Blueprint
    local pos = self:GetPosition()
    local mass = bp.Economy.BuildCostMass * (bp.Wreckage.MassMult or 0)
    local energy = bp.Economy.BuildCostEnergy * (bp.Wreckage.EnergyMult or 0)
    local time = (bp.Wreckage.ReclaimTimeMultiplier or 1)

    local prop = CreateProp( pos, wreck )

    prop:SetScale(bp.Display.UniformScale)
    prop:SetOrientation(self:GetOrientation(), true)
    prop:SetPropCollision('Box', bp.CollisionOffsetX, bp.CollisionOffsetY, bp.CollisionOffsetZ, bp.SizeX* 0.5, bp.SizeY* 0.5, bp.SizeZ * 0.5)
    prop:SetMaxReclaimValues(time, time, mass, energy)

    mass = (mass - (mass * (overkillRatio or 1))) * self:GetFractionComplete()
    energy = (energy - (energy * (overkillRatio or 1))) * self:GetFractionComplete()
    time = time - (time * (overkillRatio or 1))

    prop:SetReclaimValues(time, time, mass, energy)
    prop:SetMaxHealth(bp.Defense.Health)
    prop:SetHealth(self, bp.Defense.Health * (bp.Wreckage.HealthMult or 1))

    if not bp.Wreckage.UseCustomMesh then
        prop:SetMesh(bp.Display.MeshBlueprintWrecked)
    end

    TryCopyPose(self,prop,false)

    prop.AssociatedBP = self:GetBlueprint().BlueprintId

    return prop
end

local function DeathThread(self, overkillRatio, instigator)
    --CreateScaledBoom(self, overkillRatio)
    local bp = self:GetBlueprint()
    local army = self:GetArmy()
    local pos = self:GetPosition()
    local seafloor = GetTerrainHeight(pos[1], pos[3]) + GetTerrainTypeOffset(pos[1], pos[3])

    -- some cleanup borrowed from Unit:DeathThread
    self:DestroyAllDamageEffects()

    -- pretty debris if we're on the surface, otherwise just an explosion
    if self.PlayDestructionEffects then
        if self:GetCurrentLayer() == "Water" then
            self:CreateDestructionEffects( self, overkillRatio )
        else
            CreateScaledBoom(self, overkillRatio, 0)
        end
    end

    -- now we decide. If we have a death animation, we just wait for that to deposite us
    -- at the seafloor. If we don't, use a slider instead.
    if bp.Display.AnimationDeath then
        -- wait for it to hit the seafloor naturally, then turn us into wreckage
        local watchBone = bp.WatchBone or 0
        local seafloor = GetTerrainHeight(pos[1], pos[3]) + GetTerrainTypeOffset(pos[1], pos[3])
        --LOG(self, " sinking from ", pos[2], " to ", seafloor)
        local IsMoving = true
        local LastPos= 0
        local Timer = 0
        while self:GetPosition(watchBone)[2] > seafloor and IsMoving do
            WaitSeconds(0.1)
            Timer = Timer + 1
            if (LastPos != self:GetPosition(watchBone)[2]) then
                Timer = 0
            end
            if Timer == 10 then
                IsMoving = false
            end
            LastPos = self:GetPosition(watchBone)[2]
        end
        CreateScaledBoom(self, overkillRatio, watchBone)
        --LOG(self, " sunk!")
        CreateWreckageProp(self, overkillRatio)
    else
        -- if we're sliding, also fork a thread to create sinking effects
        self:ForkThread(function()
            local i = 0
            while true do
                local numBones = self:GetBoneCount()-1
                local sx, sy, sz = self:GetUnitSizes()
                local vol = sx * sy * sz
                local rx, ry, rz = self:GetRandomOffset(0.25)
                local rs = Random(vol/2, vol*2) / (vol*2)
                local randBone = Util.GetRandomInt( 0, numBones)

                CreateEmitterAtBone( self, randBone, army, '/effects/emitters/destruction_underwater_explosion_flash_01_emit.bp')
                :ScaleEmitter(sx)
                :OffsetEmitter(rx, ry, rz)
                CreateEmitterAtBone( self, randBone, army, '/effects/emitters/destruction_underwater_sinking_wash_01_emit.bp')
                :ScaleEmitter(sx/3)
                :OffsetEmitter(rx, ry, rz)
                CreateEmitterAtBone( self, 0, army, '/effects/emitters/destruction_underwater_sinking_wash_01_emit.bp')
                :ScaleEmitter(sx/2)
                :OffsetEmitter(rx, ry, rz)

                WaitSeconds(Util.GetRandomFloat( 0.4+i, 1.0+i))
                i = i + 0.3
            end
        end)

        local slider = CreateSlider(self, 0)
        --LOG(self, " sliding from ", pos[2], " to ", seafloor)
        slider:SetGoal(0, (seafloor - pos[2])*20 +5, 0)
        slider:SetSpeed(8)
        WaitFor(slider)
        slider:Destroy()

        CreateScaledBoom(self, overkillRatio)
        CreateWreckageProp(self, overkillRatio)
        --LOG(self, " sunk!")
    end
    self:Destroy()
end
SeaUnit = Class(oldSeaUnit) {
    DeathThread = DeathThread;
}
SubUnit = Class(oldSubUnit) {
    DeathThread = DeathThread;
}

    ##########################################################################################
    ## Submarine Running Mode Management
    ########################################################################################## 

do
local OldSubUnit = SubUnit
SubUnit = Class(OldSubUnit) {
    -- Put stealth code here for all attack subs.
    UseRunningModes = false,
    SilentRunningMode = false,
    SurfaceMode = false,
    --FiredRecently = false,
    --MovedRecently = false,
    OnStopBeingBuilt =function(self, builder, layer)
        OldSubUnit.OnStopBeingBuilt(self, builder, layer)
        --LOG("|| Done being built")
        local bp = self:GetBlueprint()
        if bp.General.UseRunningModes and bp.General.UseRunningModes == true then
            self.UseRunningModes = true
            self:DisableUnitIntel('Cloak')
            self:DisableUnitIntel('SonarStealth')
            self:SetIntelRadius('Radar', 0)
            self:DisableUnitIntel('Radar')
        end
    end,
    OnMotionHorzEventChange = function(self, new, old)
        OldSubUnit.OnMotionHorzEventChange(self, new, old)
        if self.UseRunningModes == true then
            if old == 'Stopped' then
                --LOG("|| No longer stopped")
                if self.SilentRunningMode == false then
                    if self.SurfaceMode == false then
                        self:SetIntelRadius('Sonar', self:GetBlueprint().Intel.SonarRadius * 0.45)
                    end
                    --[[
                    if self.MoveThread then
                        self.MoveThread:Destroy()
                        self.MoveThread = nil
                    end
                    self:DisableUnitIntel('SonarStealth')
                    self.MovedRecently = true
                    --]]
                elseif self.SurfaceMode == true then
                    self:SetIntelRadius('Sonar', 0)
                    self:DisableUnitIntel('Sonar')
                end
            elseif new == 'Stopped' then
                --LOG("|| Just Stopped")
                if self.SilentRunningMode == false then
                    if self.SurfaceMode == false then
                        self:SetIntelRadius('Sonar', self:GetBlueprint().Intel.SonarRadius)
                    end
                    --self.MoveThread = self:ForkThread(self.Wait30AfterStop,self)
                elseif self.SurfaceMode == true then
                    self:SetIntelRadius('Sonar', self:GetBlueprint().Intel.SonarRadius * .5)
                    self:EnableUnitIntel('Sonar')
                end
            end
        end
    end,
    OnTorpedoFire = function(self)
        --LOG("|| Torpedo Fired")
        if self.UseRunningModes == true then
            if self.FireThread then
                --LOG("|| Destroying Fire Thread")
                self.FireThread:Destroy()
                self.FireThread = nil
            end
            --LOG("|| Waiting 15 seconds in FireThread")
            self.FireThread = self:ForkThread(self.Wait15AfterFire,self)
        end
    end,
    Wait15AfterFire = function(self)
        self:RemoveToggleCap('RULEUTC_IntelToggle')
        WaitSeconds(15)
        if self.SurfaceMode == false then
            self:AddToggleCap('RULEUTC_IntelToggle')
        end
    end,

    OnScriptBitSet = function(self, bit)                                        -- Silent
        if self.UseRunningModes == true and bit == 3 then # intel toggle
            self.SilentRunningMode = true
            self:RemoveCommandCap('RULEUCC_Dive')
            self:SetSpeedMult(0.5)
            self:SetWeaponEnabledByLabel('Torpedo01', false)
            self:SetIntelRadius('Vision',10)
            self:EnableUnitIntel('SonarStealth')
            self:EnableUnitIntel('Cloak')
            self:StopUnitAmbientSound( 'ActiveLoop' )
            self:SetMaintenanceConsumptionActive()
        else
            OldSubUnit.OnScriptBitSet(self, bit)
        end
    end,

    OnScriptBitClear = function(self, bit)                                      -- Submerged, but not Silent
        if self.UseRunningModes == true and bit == 3 then # intel toggle
            self.SilentRunningMode = false
            self:AddCommandCap('RULEUCC_Dive')
            self:SetSpeedMult(1)
            self:SetWeaponEnabledByLabel('Torpedo01', true)
            local bpIntel = self:GetBlueprint().Intel
            self:SetIntelRadius('Vision', bpIntel.VisionRadius)
            self:DisableUnitIntel('SonarStealth')
            self:DisableUnitIntel('Cloak')
            if self.Stopped == true then
                self:SetIntelRadius('Sonar', bpIntel.SonarRadius)
            else
                self:SetIntelRadius('Sonar', bpIntel.SonarRadius * 0.45)
            end
            self:PlayUnitAmbientSound( 'ActiveLoop' )
            self:SetMaintenanceConsumptionInactive()
        else
            OldSubUnit.OnScriptBitClear(self, bit)
        end
    end,

    OnLayerChange = function(self, new, old)
        if self.UseRunningModes == true then
            if old == "Sub" and new == "Water" then                             -- Surfaced
                self.SurfaceMode = true
                self:RemoveToggleCap('RULEUTC_IntelToggle')
                self:SetSpeedMult(1.3)
                local bpIntel = self:GetBlueprint().Intel
                self:SetIntelRadius('Vision', bpIntel.VisionRadius * 1.5)
                self:SetIntelRadius('Sonar', bpIntel.SonarRadius * .5)
                self:SetIntelRadius('Radar', bpIntel.RadarRadius)
                self:EnableUnitIntel('Radar')
            end
            if old == "Water" and new == "Sub" then                             -- Submerged, but not Silent
                self.SurfaceMode = false
                self:AddToggleCap('RULEUTC_IntelToggle')
                self:SetSpeedMult(1)
                local bpIntel = self:GetBlueprint().Intel
                self:SetIntelRadius('Vision', bpIntel.VisionRadius)
                self:SetIntelRadius('Sonar', bpIntel.SonarRadius)
                self:SetIntelRadius('Radar', 0)
                self:DisableUnitIntel('Radar')
            end
        end
        OldSubUnit.OnLayerChange(self, new, old)
    end,
}
end


do
local OldUnit = SeaUnit
SeaUnit = Class(OldUnit) {

    InWater = true,
    SpeedFactor = 1,
    TurnFactor = 1,
    ChangedCourseRecently = false,
    InMotion = false,
    OnStopBeingBuilt = function(self, builder, layer)
        OldUnit.OnStopBeingBuilt(self, builder, layer) 
        self:ForkThread(self.SpeedManager, self) 
        --LOG("|| Ship Built")
    end, 

    OnMotionHorzEventChange = function( self, new, old )
        OldUnit.OnMotionHorzEventChange(self, new, old)
        if self.InWater == true then
            if new == 'Stopped' then
                self.InMotion = false
                self.SpeedFactor = 1
                self.TurnFactor = 1
                --LOG("|| Setting to 1")
                self:SetSpeedMult(1)

                --self:EnableUnitIntel('SonarStealth')    -- Tech 1 and 2 ships have sonar stealth when not moving.

            elseif new == 'TopSpeed' then
                self.InMotion = true
                self.ChangedCourseRecently = false

                self:DisableUnitIntel('SonarStealth')

            end
        end
    end,



--[[
        SpeedManager will run continuously, monitoring the movement of the ship and adjusting the SpeedMultiplier accordingly.
        If a ship goes straight for an extended period, its SpeedMultiplier will increase steadily.  If a turn happens, the ship
        will slow down, and lose its momentum.  Try and make this smooth.  Not sure how to handle stopping, possibly with a very
        slow brake speed.
--]]
    SpeedManager = function(self)
        local LastHeading = self:GetHeading()
        while not self:IsDead() do 
            WaitSeconds(1)
            if self.InWater == true then
                local CurrentHeading = self:GetHeading()
                if math.abs(CurrentHeading - LastHeading) > 0.05 then
                    self.ChangedCourseRecently = true
                else
                    self.ChangedCourseRecently = false
                end
                LastHeading = CurrentHeading
                if self.InMotion then
                    if self.ChangedCourseRecently == false then
                        if self.SpeedFactor < 7 then
                            self.SpeedFactor = math.min(self.SpeedFactor + .1,7)
                            self.TurnFactor = 1/self.SpeedFactor
                            self:SetSpeedMult(self.SpeedFactor)
                            --if self.SpeedFactor >= 5 then
                            --    self:SetTurnMult(.05)       -- The 2 is a multiplier that scales up the global turn rate when slow.
                            --else
                                self:SetTurnMult(self.TurnFactor * 2)       -- The 2 is a multiplier that scales up the global turn rate when slow.
                            --end
                        end
                    else
                        if self.SpeedFactor > 1 then
                            self.SpeedFactor = math.max(self.SpeedFactor - .5,1)
                            self.TurnFactor = 1/self.SpeedFactor
                            self:SetSpeedMult(self.SpeedFactor)
                            --if self.SpeedFactor >= 5 then
                            --    self:SetTurnMult(.05)       -- The 2 is a multiplier that scales up the global turn rate when slow.
                            --else
                                self:SetTurnMult(self.TurnFactor * 2)       -- The 2 is a multiplier that scales up the global turn rate when slow.
                            --end
                        end
                    end
                end
            end
        end
    end,
}
end

do
local OldUnit = AirUnit
AirUnit = Class(OldUnit)
{
    CreateUnitAirDestructionEffects = function( self, scale )
        if math.random(10) < 5 then
            CreateAttachedEmitter( self, math.random(0,self:GetBoneCount()-1), self:GetArmy(), '/effects/emitters/flamethrower_emit.bp'):ScaleEmitter(.3)
        end
        OldUnit.CreateUnitAirDestructionEffects(self, scale)
    end,


    #Afterburn function
    OnDamage = function(self, instigator, amount, vector, damageType)
        OldUnit.OnDamage(self, instigator, amount, vector, damageType)
        if self:GetBlueprint().Air.HasAfterburner then
            ##-------------------------------##
            ##  Initialze Global Variables   ##   ONLY DONE ON FIRST TIME!
            ##-------------------------------##
            
            ## numOfHits (Number of Times Hit since last Afterburn) ##
            if self.numOfHits == nil then
               self.numOfHits = 0
            end

            ## dmgAccumlated (Amt of dmg sustained since last Afterburn) ##
            if self.dmgAccumulated == nil then
               self.dmgAccumulated = 0
            end

            ##--------------------------------##
            ##  Update the Global Variables   ##   DONE EVERY TIME !!
            ##--------------------------------##
            self.numOfHits = self.numOfHits + 1
            self.dmgAccumulated = self.dmgAccumulated + amount

            ##------------------------------------------------------------------------------------##
            ## DO ONLY if (Plane has Fuel) AND (Not Afterburning Now) AND (Meets Damage Criteria) ##
            ##------------------------------------------------------------------------------------##
            if self.Thread == nil and self:GetFuelRatio() > 0  then
                if self.dmgAccumulated >= (self:GetBlueprint().Defense.MaxHealth / 10) or self.numOfHits >= 6 then

                    ##---------------------------##
                    ## Reset my Global Variables ##
                    ##---------------------------##
                    self.numOfHits = 0
                    self.dmgAccumulated = 0
            
                    ##----------------------##
                    ## Speed/Acc/Turn Boost ##
                    ##----------------------##

                    self:SetSpeedMult(3)
                    self:SetAccMult(3)
                    self:SetTurnMult(3)

                    ##----------------------------##
                    ## Afterburner Exhaust Effect ##
                    ##----------------------------##

                    CreateAttachedEmitter( self, -1, self:GetArmy(), '/effects/emitters/dirty_exhaust_smoke_01_emit.bp')
--                    CreateAttachedEmitter( self, 'Exhaust_Right', self:GetArmy(), '/effects/emitters/dirty_exhaust_smoke_01_emit.bp')

                    ##------------------------------##
                    ## Revert Stats and Update Fuel ##
                    ##------------------------------##

                    updateSpeedAndFuel = function(self)
                        local preBurnFuel = self:GetFuelRatio()
                        WaitSeconds(2)
                        if not self:IsDead() then
                            local postBurnFuel = self:GetFuelRatio()

                            local newFuelRatio = preBurnFuel - (3*(preBurnFuel - postBurnFuel))

                            if newFuelRatio > 0 then
                                self:SetSpeedMult(1)
                                self:SetAccMult(1)
                                self:SetTurnMult(1)
                                self:SetFuelRatio(newFuelRatio)
                            else
                                self:SetSpeedMult(0.25)
                                self:SetAccMult(0.25)
                                self:SetTurnMult(0.25)
                                self:SetFuelRatio(0)
                                self:OnRunOutOfFuel(self)
                            end
                        end

                        WaitSeconds(1)
                        self.Thread = nil     ##Permit next afterburn
                    end
                    ##---------------------------##
                    ## Execute the Actual Update ##
                    ##---------------------------##
                    self.Thread = self:ForkThread(updateSpeedAndFuel)
                end
            end
        end
    end,
}
end


#-------------------------------------------------------------
#  MASS FABRICATION UNITS
#-------------------------------------------------------------
do
local OldUnit = MassFabricationUnit
MassFabricationUnit = Class(OldUnit)
{
    UnderAttack = false,
    DeathWeaponModifier = 1,
    EnergyUsage = 0,
    On = true,
    OnStopBeingBuilt = function(self, builder, layer)
        OldUnit.OnStopBeingBuilt(self, builder, layer) 
        self:ForkThread(self.FabricationManagementThread, self) 
        self:ForkThread(self.AttackWatchThread, self) 
        local EconBP = self:GetBlueprint().Economy
        self.EnergyUsage = EconBP.MaintenanceConsumptionPerSecondEnergy
        if self:GetCurrentLayer() == 'Seabed' then
            local UnderwaterProductionFactor = EconBP.UnderwaterProductionFactor or 1
            self:SetProductionPerSecondMass(EconBP.ProductionPerSecondMass * UnderwaterProductionFactor)
        end
    end, 
    OnDamage = function(self, instigator, amount, vector, damageType)
        self.UnderAttack = true
        OldUnit.OnDamage(self, instigator, amount, vector, damageType)
    end, 
    DoDeathWeapon = function(self)
        if self:IsBeingBuilt() then return end
        local bp = self:GetBlueprint()
        for k, v in bp.Weapon do
            if(v.Label == 'DeathWeapon') then
                if v.FireOnDeath == true then
                    self:SetWeaponEnabledByLabel('DeathWeapon', true)
                    self:GetWeaponByLabel('DeathWeapon'):Fire()
                else
                    self:ForkThread(self.DeathWeaponDamageThread, v.DamageRadius, v.Damage * self.DeathWeaponModifier, v.DamageType, v.DamageFriendly, v.EdgeDamageRatio)
                end
                break
            end
        end
    end,
    AttackWatchThread = function(self) 
        while not self:IsDead() do
            if self.UnderAttack == true then
                WaitSeconds(5) 
                if self:GetScriptBit('RULEUTC_ProductionToggle') == false then 
                    self:SetScriptBit('RULEUTC_ProductionToggle', true) 
                    self:SetArmoredState()
                    self.DeathWeaponModifier = .1
                end 
                WaitSeconds(15) 
                self.UnderAttack = false
            end
            WaitSeconds(1) 
        end 
    end, 
    FabricationManagementThread = function(self)   
         local DesiredStoragePercentage = 0.75   
         while not self:IsDead() do   
             if not self.UnderAttack then   
                 local aiBrain = self:GetAIBrain()   
                 local MassStorageRatio = aiBrain:GetEconomyStoredRatio('MASS')   
                 local EnergyStorageRatio = aiBrain:GetEconomyStoredRatio('ENERGY')   
                 local EnergyTrend = aiBrain:GetEconomyTrend('ENERGY') * 10   
                 local EnergyStored = aiBrain:GetEconomyStored('ENERGY')   
    
                 #Convert percentage into actual value.   
                 local DesiredStorageAmount = 1 / EnergyStorageRatio * EnergyStored * DesiredStoragePercentage   
                 #LOG("*debug: DesiredStorage: " .. DesiredStorageAmount)   
    
    
                 if MassStorageRatio == 1 then   
                     if EnergyStorageRatio < 1 then   
                         self:SwitchOff()   
                     end   
                 else   
                     -- If my energy would last more than 10 seconds if I turned this on.   
                      if self.On == false and (EnergyStored + ((EnergyTrend - self.EnergyUsage) * 10)) > DesiredStorageAmount then
                         self:SwitchOn()   
                     -- If my energy stored would last less than 5 seconds if I stay on, turn off.   
                     elseif self.On == true and (EnergyStored + (EnergyTrend * 5) < DesiredStorageAmount) then   
                         self:SwitchOff()   
                     end   
                 end   
             end   
             WaitSeconds(2)   
         end   
     end,   
--[[
    FabricationManagementThread = function(self) 
        while not self:IsDead() do
            if not self.UnderAttack then
                local aiBrain = self:GetAIBrain() 
                local MassStorageRatio = aiBrain:GetEconomyStoredRatio('MASS')
                local EnergyStorageRatio = aiBrain:GetEconomyStoredRatio('ENERGY')
                local EnergyTrend = aiBrain:GetEconomyTrend('ENERGY') * 10
                local EnergyStored = aiBrain:GetEconomyStored('ENERGY')
                if MassStorageRatio == 1 then
                    if EnergyStorageRatio < 1 then
                        self:SwitchOff()
                    end
                else
                    -- If energy stored would last longer than 10 seconds after turning on this fabricator, then switch to On.
                    if self.On == false and EnergyStored + ((EnergyTrend - self.EnergyUsage) * 10) > 0 then
                        self:SwitchOn()
                    -- If energy stored would last less than 5 seconds if we keep this fabricator on, then turn on.
                    elseif self.On == true and EnergyStored + (EnergyTrend * 5) < 0 then
                        self:SwitchOff()
                    end
                end
            end
            WaitSeconds(2)
        end 
    end, 
--]]
SwitchOff = function(self)
        if self:GetScriptBit('RULEUTC_ProductionToggle') == false then 
            self:SetScriptBit('RULEUTC_ProductionToggle', true) 
            self:SetArmoredState()
            self.DeathWeaponModifier = .1
            self.On = false
        end 
    end,
    SwitchOn = function(self)
        if self:GetScriptBit('RULEUTC_ProductionToggle') == true then 
            self:SetScriptBit('RULEUTC_ProductionToggle', false) 
            self:SetUnarmoredState()
            self.DeathWeaponModifier = 1
            self.On = true
        end 
    end,
}
end


--[[
    ##########################################################################################
    ## SHIELD UNITS
    ##########################################################################################
do
local OldUnit = ShieldStructureUnit
ShieldStructureUnit = Class(OldUnit) {
    TrackingDamage = false,
    TrackedDamage = 0,
    Sputtering = false,
    DoTakeDamage = function(self, instigator, amount, vector, damageType)
        OldUnit.DoTakeDamage(self, instigator, amount, vector, damageType)
        self.TrackedDamage = self.TrackedDamage + amount
        if not self.TrackingDamage then
            self:ForkThread(self.TrackDamage,self)
        end
        if not self.Sputtering then
            -- If shield unit itself accumulates an amount of damage equal to least 25% of
            -- its max health in one second, turn off shield for 1 second.
            if self.TrackedDamage > self:GetBlueprint().Defense.MaxHealth / 4 then
                self:ForkThread(self.Sputter,self)
            end
        end
    end,
    TrackDamage = function(self)
LOG("|| Tracking Damage")
        self.TrackingDamage = true
        WaitSeconds(1)
        self.TrackedDamage = 0
        self.TrackingDamage = false
LOG("|| Done Tracking Damage")
    end,
    Sputter = function(self)
        self:StopUnitAmbientSound( 'ActiveLoop' )
        self:DisableShield()
        self.Sputtering = true
        WaitSeconds(1)
        self:PlayUnitAmbientSound( 'ActiveLoop' )
        self:EnableShield()
        WaitSeconds(0.5)
        self.Sputtering = false
    end,
}
end
--]]


end
